// Buffer remove operator class -*- c++ -*-

#ifndef BUFFERREMOVE_H_
# define BUFFERREMOVE_H_
# pragma interface

# include "Expression.h"

/** @file BufferRemove.h
 * Operation for removing queue or stack items
 */

/** Buffer remove operator expression */
class BufferRemove : public Expression
{
public:
  /** Constructor
   * @param buffer	Buffer whose items are being removed
   * @param amount	Number of items to remove (NULL=1)
   * @param index	Indexing expression (optional)
   */
  BufferRemove (class Expression& buffer,
		class Expression* amount,
		class Expression* index);

private:
  /** Copy constructor */
  BufferRemove (const class BufferRemove& old);
  /** Assignment operator */
  class BufferRemove& operator= (const class BufferRemove& old);
protected:
  /** Destructor */
  ~BufferRemove ();
public:

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eBufferRemove; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  /** Equality comparison operator */
  bool operator== (const class BufferRemove& other) const {
    return
      (myAmount || !other.myAmount) && (other.myAmount || !myAmount) &&
      (myIndex || !other.myIndex) && (other.myIndex || !myIndex) &&
      *myBuffer == *other.myBuffer &&
      (!myAmount || *myAmount == *other.myAmount) &&
      (!myIndex || *myIndex == *other.myIndex);
  }

  /** Ordering comparison operator */
  bool operator< (const class BufferRemove& other) const {
    if (!myAmount && other.myAmount)
      return true;
    if (!other.myAmount && myAmount)
      return false;
    if (!myIndex && other.myIndex)
      return true;
    if (!other.myIndex && myIndex)
      return false;
    if (myAmount) {
      if (*myAmount < *other.myAmount) return true;
      if (*other.myAmount < *myAmount) return false;
    }
    if (myIndex) {
      if (*myIndex < *other.myIndex) return true;
      if (*other.myIndex < *myIndex) return false;
    }
    return *myBuffer < *other.myBuffer;
  }

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation& valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @param transition	Transition for registering quantified variables
   * @param declare	flag: declare new variables if required
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation,
			    class Transition* transition,
			    bool declare);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

  /** Determine whether the expression depends on a set of variables
   * @param vars	the set of variables
   * @param complement	flag: treat the set as its complement
   */
  bool depends (const class VariableSet& vars,
		bool complement) const;

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @param data	parameters to be passed to the operation
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class Expression&,void*),
		     void* data) const;

# ifdef EXPR_COMPILE
  /** Generate C code for evaluating the expression
   * @param cexpr	the compilation
   * @param indent	indentation level
   * @param lvalue	C expression referring to the lvalue
   * @param vars	the variables that have been assigned a value
   */
  void compile (class CExpression& cexpr,
		unsigned indent,
		const char* lvalue,
		const class VariableSet* vars) const;
# endif // EXPR_COMPILE

  /** Display the expression
   * @param printer	the printer object
   */
  void display (const class Printer& printer) const;

private:
  /** The buffer expression */
  class Expression* myBuffer;
  /** The optional amount expression */
  class Expression* myAmount;
  /** The optional index expression */
  class Expression* myIndex;
};

#endif // BUFFERREMOVE_H_
