// undefined expression class -*- c++ -*-

#ifndef UNDEFINED_H_
# define UNDEFINED_H_
# pragma interface

# include "Expression.h"

/** @file Undefined.h
 * Undefined literal
 */

/** Undefined expression */
class Undefined : public Expression
{
public:
  /** Fatalness levels */
  enum Fatalness {
    /** issue an error message; do not fire the transition instance */
    fError,
    /** issue a fatal error message; stop the graph generation */
    fFatalError
  };

  /** Constructor
   * @param fatalness	Fatalness level
   */
  explicit Undefined (enum Fatalness fatalness);

private:
  /** Copy constructor */
  explicit Undefined (const class Undefined& old);
  /** Assignment operator */
  class Undefined& operator= (const class Undefined& old);
protected:
  /** Destructor */
  ~Undefined ();
public:

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eUndefined; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  /** Equality comparison operator */
  bool operator== (const class Undefined& other) const {
    return myFatalness == other.myFatalness;
  }

  /** Ordering comparison operator */
  bool operator< (const class Undefined& other) const {
    return myFatalness < other.myFatalness;
  }

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation& valuation) const;

  /** Partially evaluate the expression using a valuation
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation&,
			    class Transition*, bool) { return copy (); }

  /** Substitute some variables in the expression with expressions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution&) { return copy (); }

  /** Determine whether the expression depends on a set of variables
   * @param vars	the set of variables
   * @param complement	flag: treat the set as its complement
   */
  bool depends (const class VariableSet& vars,
		bool complement) const;

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @param data	parameters to be passed to the operation
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class Expression&,void*),
		     void* data) const;

# ifdef EXPR_COMPILE
  /** Generate C code for evaluating the expression
   * @param cexpr	the compilation
   * @param indent	indentation level
   * @param lvalue	C expression referring to the lvalue
   * @param vars	the variables that have been assigned a value
   */
  void compile (class CExpression& cexpr,
		unsigned indent,
		const char* lvalue,
		const class VariableSet* vars) const;
# endif // EXPR_COMPILE

  /** Display the expression
   * @param printer	the printer object
   */
  void display (const class Printer& printer) const;

private:
  /** Fatalness level */
  enum Fatalness myFatalness;
};

#endif // UNDEFINED_H_
