// Probabilistic reachability set storage -*- c++ -*-

#ifndef HASHGRAPH_H_
# define HASHGRAPH_H_
# pragma interface

# include "BitBuffer.h"

/** @file HashGraph.h
 * Transient, probabilistic reachability set storage
 */

/** Probabilistic reachability set storage */
class HashGraph
{
public:
  /** Constructor */
  HashGraph () :
    myNumTables (0), myFuncSize (0), myHashSize (0),
    myHashers (0), myHashes (0), myNumStates (0), myNumArcs (0) {}
private:
  /** Copy constructor */
  HashGraph (const class HashGraph& old);
  /** Assignment operator */
  class HashGraph& operator= (const class HashGraph& old);
public:
  /** Destructor */
  ~HashGraph ();

  /** Get the number of states */
  card_t getNumStates () const { return myNumStates; }
  /** Get the number of arcs */
  card_t getNumArcs () const { return myNumArcs - 1; }

  /** Initialize the hash tables
   * @param numTables	number of hash tables
   * @param funcSize	(in/out) number of elements in the hash functions
   * @param hashSize	(in/out) number of elements in the hash tables
   * @return		true if the initialization succeeded
   */
  bool init (size_t numTables,
	     size_t& funcSize,
	     size_t& hashSize);

  /** Add a state to the graph
   * @param state	the encoded state
   * @param size	length of the encoded state, in bytes
   * @return		0 if the state exists; otherwise, a new state number
   */
  card_t add (const void* state,
	      size_t size);

private:
  /** Number of hash tables */
  size_t myNumTables;
  /** Number of elements in the hash function tables (power of 2) */
  size_t myFuncSize;
  /** Number of elements in the hash tables (power of 2) */
  size_t myHashSize;
  /** The hash functions */
  size_t** myHashers;
  /** The hash tables */
  size_t** myHashes;
  /** Number of generated states */
  card_t myNumStates;
  /** Number of generated arcs (invocations to add ()) */
  card_t myNumArcs;
};

#endif // HASHGRAPH_H_
